/*
 * This program is Copyright 1998 by Henry Cohen commercial rights are reserved.
 * The program is provided as is with no warranties of any sort including 
 * fitness for any purpose.  In no case will the author be held liable for
 * damages related to the use of this program.

 * The author hereby grants rights to use, copy, modify and create derived
 * products for individual, non-commercial purposes provided a copy of this
 * notice is included.  Any commercial use requires the written approval of 
 * the author.
 * Henry Cohen 1998.
 */

/**
 * Class JJDic - top level class for Jdic application.
 * This class has the main function that is used to invoke the
 * whole application.  This class ties together all of the pieces.
 * This would be different for an applet.  I am going to do things in
 * here that would not be allowed in an applet.
 */
package jdic;

import java.io.*;
import java.util.*;


public class JJDic {
    /*
     * Private constants
     */
    private final static int MAX_DICTIONARIES = 10;
    
    /*
     * Class variables.
     */
    static GeneralDictionary[] dictionaries;
    static boolean[] dictionaryEnabled;
    static String[] dictionaryName;
    static int ndicts = 0;       // number of loaded dictionaries
    static KanjiDictionary kanjiDic;  // Kanjidic is a special case
    static boolean showDictNames = true;
    
    static Gui mainGui;
    
    /*
     * Private fields.  Instance variables.
     * None this class has only one instance and all fields are
     * static.
     */
    static Runtime runtime;
   

    public static void main(String[] argv){    
	/*
	 * Get Default properties, System then User.
	 *[soap] it looks like I can only support user properties not
	 * global application properties.  The reason being that there
	 * does not seem to be a 100% pure way to get a standard file
	 * location other than relative to the user's home directory or
	 * to the Java installation directory.  Now if we could install
	 * application.defaults somewhere relative to JavaHome then it
	 * might be doable. 
	 */
	runtime = Runtime.getRuntime();
	dictionaries = new GeneralDictionary[MAX_DICTIONARIES];
	dictionaryEnabled = new boolean[MAX_DICTIONARIES];
	dictionaryName = new String[MAX_DICTIONARIES];
	String dictPath = null, kanjiPath = null;

	String home = new String(System.getProperty("user.home"));

	File usrPropFile = new File(home, ".jjdic");
	FileInputStream propStream;
	Properties userProperties = new Properties();
		
	try {
	    propStream = new FileInputStream(usrPropFile);	
	    userProperties.load(propStream);
	    propStream.close();		
	    dictPath = userProperties.getProperty("Dictionaries");
	    kanjiPath = userProperties.getProperty("Kanjidic");
	} catch (FileNotFoundException e){
	    System.err.println(e);
	} catch (IOException e) {
	    System.err.println(e);
	    // No properties file. Punt. 
	}
	runtime.gc();
	
	mainGui = new Gui();
	if (kanjiPath != null) {
	    initKanjiDic(kanjiPath);
	} else {
	    kanjiPath = mainGui.getKanjiPath();
	    initKanjiDic(kanjiPath);
	}
	runtime.gc();
	if (dictPath != null) {
	    initDicts(dictPath);
	}
	
    }
    /*
     * parse a property list dictionary entry and call loadNewDictionary
     * to load the dictionaries.  A dictionary list is a string consisting
     * of paths to dictionaries seperated by the system path seperator character
     * each element can optionally have a prefix specifying the encoding of
     * the dictionary.  If no encoding specification is present EUCJIS is 
     * assumed. e.g. for Windoze "c:\jdic\edict;SJIS,c:\jdic\privatedic"
     * Kanjidic should not be in this list, it is handled seperately.
     */
    private static void initDicts(String pathList) {
	String path, encoding;
	char pathsep = System.getProperty("path.separator").charAt(0);
	int comma;
	for (int next = pathList.indexOf(pathsep); next != -1; 
	     pathList = pathList.substring(next+1),next = pathList.indexOf(pathsep)) {
	    path = pathList.substring(0, next);
	    comma = path.indexOf(',');
	    if (comma != -1) {
		encoding = path.substring(0,comma);
		path = path.substring(comma+1);
  		loadNewDictionary(path, encoding, true);
	    } else {
 		loadNewDictionary(path, "EUCJIS", true);		
	    }
	runtime.gc();
	}
	/*
	 * Last or only element
	 */
	path = pathList;
	comma = path.indexOf(',');
	if (comma != -1) {
	    encoding = path.substring(0,comma);
	    path = path.substring(comma+1);
 	    loadNewDictionary(path, encoding, true);
	} else {
  	    loadNewDictionary(path, "EUCJIS", true);		
	}
    }
    /**
     * method loadNewDictionary(directory, file) loads a dictionary from directory/file.
     * Called during startup for dictionaries in the user's properties file 
     * of from the gui callback routine for open menu item.
     */
    static void loadNewDictionary(String path, String encoding, boolean mkindx){
	mainGui.setStatus("Loading dictionary"+path);
	
	try {
	    dictionaries[ndicts] = new GeneralDictionary(new File(path), mkindx, encoding);
	} catch (BadIndexException e) {
	    mainGui.setStatus("load failed. "+e);
	    return;
	} catch (EOFException e){
	    mainGui.setStatus("load failed. "+e);
	    return;
	} catch (UnsupportedEncodingException e){
	    mainGui.setStatus("load failed. "+e);
	    return;
	} catch (FileNotFoundException e){
	    mainGui.setStatus("load failed. "+e);
	    return;
	} catch (IOException e){
	    mainGui.setStatus("load failed. "+e);
	    return;
	}
	dictionaryEnabled[ndicts] = true;
	dictionaryName[ndicts] = mainGui.addDictMenu(path);
	ndicts++;
	mainGui.setStatus("dictionary loaded");
    }
    static void setDictionaryState(String dname, boolean dstate) {
	for (int i = 0; i<ndicts; i++) {
	    if (dictionaryName[i].equals(dname)) {
		dictionaryEnabled[i] = dstate;
		break;
	    }
	}
    }
    static void initKanjiDic(String path) {
	mainGui.setStatus("Loading Kanji dictionary"+path);
	int comma;
	String encoding;
	
	comma = path.indexOf(',');
	if (comma != -1) {
	    encoding = path.substring(0,comma);
	    path = path.substring(comma+1);
	} else {
	    encoding = "EUCJIS";
	}	
	try {
	    kanjiDic = new KanjiDictionary(new File(path), true, encoding);
	} catch (BadIndexException e) {
	    mainGui.setStatus("Kanjidic load failed"+e);
	    return;
	} catch (EOFException e){
	    mainGui.setStatus("Kanjidic load failed"+e);
	    return;
	} catch (UnsupportedEncodingException e){
	    mainGui.setStatus("Kanjidic load failed"+e);
	    return;
	} catch (FileNotFoundException e){
	    mainGui.setStatus("Kanjidic load failed"+e);
	    return;
	} catch (IOException e){
	    mainGui.setStatus("Kanjidic load failed"+e);
	    return;
	}
	mainGui.kdictLoaded();
    }
    static void formQuery(String query){
	boolean found = false;
	String results[];
	String notFound[] = {"\""+query+"\""+" not found."};
	
	mainGui.setStatus("Searching for "+query);
	mainGui.clearResults();
	for (int i = 0; i<ndicts; i++) {
	    if (dictionaryEnabled[i]) {
		results = dictionaries[i].lookup(query,mainGui.firstKanji);
		found = found || results != null;
		if (showDictNames && results != null) {
		    String[] tmp = new String[results.length];
		    for (int j = 0; j< results.length; j++){
			tmp[j] = dictionaryName[i]+":"+results[j];
		    }
		    mainGui.appendResults(tmp);
		} else {
		    mainGui.appendResults(results);
		}
	    }
	}
	if (!found) {
	    mainGui.appendResults(notFound);
	}
	mainGui.setStatus("Search complete.");
    }
    /*
     * the method kdictQuery takes a string provided by the GUI and
     * parses it into words to lookup in the kanji dictionary.  There are
     * four kinds of queries, 1) simple kanji queries. We can have up to 100
     * kanji in a single query.  Kanji are just looked
     * up as they are. 2,3) On, Kun yomikata queries.  Kana words are looked
     * up as on/kun yomi.  We only allow one yomikata query at a time to keep
     * the output simple.  For any given kana string both on and kun yomi are
     * looked up. 4) English keywords:
     * romaji strings are looked up as english keywords.  The resulting strings
     * are returned to the GUI for display as keys and records. Again only one
     * english keyword search is supported at a time.  The variable names make
     * better sense if you know a little Japanese.
     */
    static void kdictQuery(String query){
	String[] tmpList = new String[100];
	String[] kanjiList;
	String[] onYomiList = null, kunYomiList = null, eigoList = null;
	int current_position = 0;
	int mark_position;
	int scan_end = query.length();
	int kanjiCount = 0;
	boolean yomikata_done=false, eigo_done=false;
		
	mainloop: while (current_position < scan_end) {
	    /*
	     * Skip any leading punctuation or whitespace
	     */
	    while (!MojiBunkai.isDictChar(query.charAt(current_position))) {
		current_position++;
		if (current_position >= scan_end) break mainloop;
	    }
	    /*
	     * lookup all kanji only one result per lookup here.
	     */
	    if (MojiBunkai.isKanji(query.charAt(current_position))) {
		String[] s = kanjiDic.lookup(query.substring(current_position,current_position+1), true);
		if (s != null)
		    tmpList[kanjiCount++] = s[0];
		current_position++;
		continue mainloop;
	    }
	    /*
	     * Katakana word?
	     */
	    if (MojiBunkai.isKatakana(query.charAt(current_position))) {
		if (!yomikata_done) {
		    mark_position = current_position;
		    try {
			while (MojiBunkai.isKatakana(query.charAt(++current_position)));
		    } catch (StringIndexOutOfBoundsException e){};
		    onYomiList = kanjiDic.lookup(query.substring(mark_position,current_position), false);
		    char[] kunBuffer = new char[current_position-mark_position];
		    for (int i = mark_position; i<current_position; i++) {
			kunBuffer[i-mark_position] = MojiBunkai.toHiragana(query.charAt(i));
		    }
		    kunYomiList = kanjiDic.lookup(new String(kunBuffer), false);
		    yomikata_done = true;
		} else {
		    try {
			while (MojiBunkai.isKana(query.charAt(current_position)))
			    current_position++;
		    } catch (StringIndexOutOfBoundsException e){};
		}
		continue mainloop;
	    }
	    /*
	     * Hiragana word?
	     */
	    if (MojiBunkai.isHiragana(query.charAt(current_position))) {
		if (!yomikata_done) {
		    mark_position = current_position;
		    try {
			while (MojiBunkai.isHiragana(query.charAt(++current_position)));
		    } catch (StringIndexOutOfBoundsException e){};
		    kunYomiList = kanjiDic.lookup(query.substring(mark_position,current_position), false);
		    char[] onBuffer = new char[current_position-mark_position];
		    for (int i = mark_position; i<current_position; i++) {
			onBuffer[i-mark_position] = MojiBunkai.toKatakana(query.charAt(i));
		    }
		    onYomiList = kanjiDic.lookup(new String(onBuffer), false);
		    yomikata_done = true;
		} else {
		    try {
			while (MojiBunkai.isKana(query.charAt(current_position)))
			    current_position++;
		    } catch (StringIndexOutOfBoundsException e){};
		}
		continue mainloop;
	    }
	    /*
	     * Romaji word 
	     */
	    if (MojiBunkai.isRomaji(query.charAt(current_position))){
		if (!eigo_done) {
		    mark_position = current_position;
		    try {
			while (MojiBunkai.isRomaji(query.charAt(++current_position)));
		    } catch (StringIndexOutOfBoundsException e){};
		    eigoList = kanjiDic.lookup(query.substring(mark_position,current_position), false);
		    eigo_done = true;
		} else {
		    try {
			while (MojiBunkai.isRomaji(query.charAt(current_position)))
			    current_position++;
		    } catch (StringIndexOutOfBoundsException e){};
		}
	    } 
	} // mainloop
	if (kanjiCount > 0) {
	    kanjiList = new String[kanjiCount];
	    for(int i = 0; i<kanjiCount;i++) kanjiList[i] = tmpList[i];
	} else {
	    kanjiList = null;
	}
	mainGui.kanjiIndexDialog(kanjiList, onYomiList, kunYomiList, eigoList);
    }
    /*
     * Do a kanjidic lookup based on radical and stroke count.
     * This search is not as accurate as I would like because this kanjidic
     * cannot distinguish between different forms of a radical with different
     * stroke counts.  I will use a vector here to hold the intermediate results.
     * I would just as soon use a bitset but I can't get the number of set bits
     * in a bitset without counting them one by one.
     */
    public static String[] doKdicRadStrokeQuery(short radicalNumber, short strokes){
	String[] radicalSet = kanjiDic.lookup("B"+radicalNumber, false);
	/* The " " is important.  It prevents matching S11 == S1 */
	String strokeString = "S"+strokes+" ";   
	Vector resultVec = new Vector();
	int index;

	for (int i = 0; i< radicalSet.length; i++){
	    index = radicalSet[i].indexOf(strokeString);
	    if (index != -1) {
		resultVec.addElement(radicalSet[i]);
	    }
	}
	int resultCount = resultVec.size();
	String[] results = new String[resultCount];
	resultVec.copyInto(results);
	return results;
    }
    static void setShowDictNames(boolean value) {
	showDictNames = value;
    }
}



