/*
 * This program is Copyright 1998 by Henry Cohen commercial rights are reserved.
 * The program is provided as is with no warranties of any sort including 
 * fitness for any purpose.  In no case will the author be held liable for
 * damages related to the use of this program.

 * The author hereby grants rights to use, copy, modify and create derived
 * products for individual, non-commercial purposes provided a copy of this
 * notice is included.  Any commercial use requires the written approval of 
 * the author.
 * Henry Cohen 1998.
 */

package jdic;
/**
 * Class KanjiInput implements a dialog to input Japanese Kanji by bushu and stroke.
 * The dialog can also generate a kdict display to show the kanji data so it can
 * be used as a bushu/stroke lookup method for kanjidic.
 */

import java.awt.*;
import java.awt.event.*;

class KanjiInput extends Dialog {
    /*
     * Instance Variables
     *
     * Variables pertaining to the whole dialog
     */
    TextComponent where;
    GridLayout grid;

    /* RadTable is a utility object. 
     * This forces loading of the RadTable class but all of it's methods and
     * fields are static so I never use the variable radTab.
     * It turns out that this is necessary because if I just use the methods
     * without calling the constructor then the static initialization of the
     * class variables will not occur.
     */
    private RadTable radTab = new RadTable(); 

    /*
     * Subdialog container panels and related variables
     */
    Panel radStrokePanel;     // Panel for radical strokes question
    Choice rStrokesChoice;    // Choice component for user input
    short radStrokes = 0;     // Answer to radical strokes question

    Panel plusStrokesPanel;   // Panel for plus how many question
    Choice plusChoice;        // Choice component for user input
    short plusStrokes = 0;    // Answer to plus how many question
    
    /*
     * Subsidiary dialogs.  In Java 1.2 these can be defined as subsidiary to
     * thisKinput.  Both will persist until dismissed by the user.  This allows
     * the user to backtrack in the lookup procedure.  In general backtracking
     * is supported at all steps.
     */
    Dialog radDialog;    // Dialog for radical selection
    Dialog kanjiDialog;  // Dialog for kanji selection
    char radChar;        // Character selected
    char displayChar;    // Displayable character for noJis radicals.
    
    /**
     * Constructor KanjiInput(TextComponent tcomp): make a kanjiInput object for
     * inputting kanji Characters into the textComponent tcomp.  The kanjiInput
     * object is persistent and will hang around until explicitly dismissed by
     * the user.  Therefore once one has been created it can be reused as often
     * as the user likes. 
     */
    /*
     * Constructor builds the KanjiInput instance. It creates and displays
     * the first dialog. The dialog is added to later as the selection
     * algorithm progresses.  The main dialog has a grid layout with a
     * single column of cells.  Each cell contains a panel and each panel
     * represents a phase of the dialog with the user.  The first panel asks
     * for the number of strokes in the radical.  A seperate kanji selection
     * panel is presented for radical selection.  When the user has selected
     * a radical the second panel is added to the main dialog showing the selected
     * radical and asking how many more strokes in the target kanji.
     */

    KanjiInput (Frame f, TextComponent tcomp) {
	super(f, "Kanji Input");
	this.addWindowListener(new WindowAdapter(){
	    public void windowClosing(WindowEvent e) {
		if(radDialog != null) radDialog.dispose();
		if (kanjiDialog != null) kanjiDialog.dispose();
		KanjiInput.this.removeAll();
		KanjiInput.this.dispose();
	    }
	});
	where = tcomp;
	grid = new GridLayout(0,1);
	this.setLayout(grid);

	radStrokePanel = new Panel();
	radStrokePanel.add(new Label("Radical Stroke Count"));
	rStrokesChoice = new Choice();
	rStrokesChoice.add("1");rStrokesChoice.add("2");rStrokesChoice.add("3");
	rStrokesChoice.add("4");rStrokesChoice.add("5");rStrokesChoice.add("6");
	rStrokesChoice.add("7");rStrokesChoice.add("8");rStrokesChoice.add("9");
	rStrokesChoice.add("10");rStrokesChoice.add("11");rStrokesChoice.add("12");
	rStrokesChoice.add("13");rStrokesChoice.add("14");rStrokesChoice.add("15");
	rStrokesChoice.add("16");rStrokesChoice.add("17");
	/*
	 * I provide two ways to answer the question.  A choice selection is
	 * enough or the user can push the OK button.  The button is there because
	 * on the first try the choice is 1 but it doesn't generate an item event
	 * unless the item state changes.  Having the button allows the user
	 * to select 1 on the first try.
	 */
	rStrokesChoice.addItemListener(new ItemListener(){
	    public void itemStateChanged(ItemEvent e) {
		short strokes = (short)Integer.parseInt((String) e.getItem());
		if (radStrokes == strokes) return;
		if (plusStrokesPanel != null) 
		    KanjiInput.this.remove(plusStrokesPanel);
		radStrokes = strokes;
		String[] radList = RadTable.glyphList(strokes);
		KanjiInput.this.radicalDialog(radList);
	    }    
	});
	radStrokePanel.add(rStrokesChoice);	
	Button goButton = new Button("Go");
	radStrokePanel.add(goButton);
	goButton.addActionListener(new ActionListener(){
	    public void actionPerformed(ActionEvent e){
		short strokes = (short)Integer.parseInt
		    (rStrokesChoice.getSelectedItem());
		if (radStrokes == strokes) return;
		if (plusStrokesPanel != null) 
		    KanjiInput.this.remove(plusStrokesPanel);
		radStrokes = strokes;
		String[] radList = RadTable.glyphList(strokes);
		KanjiInput.this.radicalDialog(radList);
	    }
	});
	add(radStrokePanel);
	if (plusStrokesPanel != null)
	    remove(plusStrokesPanel);
	validate();
	pack();
	show();
	requestFocus();
    }

    void radicalDialog(String[] glyphList) {
	if (radDialog != null) {
	    radDialog.dispose();
	}
	if (plusStrokesPanel != null)
	    remove(plusStrokesPanel);
	
	radDialog = new Dialog((Frame)this.getParent(),radStrokes+" Stroke Radicals");
	radDialog.addWindowListener(new WindowAdapter(){
	    public void windowClosing(WindowEvent e){
		radDialog.dispose();
	    }
	});
	radDialog.add(new RadicalPanel(glyphList, this), "Center");
	radDialog.pack();
	radDialog.show();
	radDialog.requestFocus();
    }

    public void addPlusStrokes(char radChar) {
	this.radChar = radChar;
	Substitute s = RadTable.radSubstitute(radChar);
	if (s != null) 
	    displayChar = s.substitute;
	else
	    displayChar = radChar;
	
	if (plusStrokesPanel == null) {
	    plusStrokesPanel = new Panel();
	} else {
	    plusStrokesPanel.removeAll();
	}
	if (kanjiDialog != null)
	    kanjiDialog.dispose();
	plusStrokesPanel.add (new Label("Radical "+displayChar+" plus "));
	plusChoice = new Choice();
	plusChoice.add("0");plusChoice.add("1");plusChoice.add("2");
	plusChoice.add("3");plusChoice.add("4");plusChoice.add("5");
	plusChoice.add("6");plusChoice.add("7");plusChoice.add("8");
	plusChoice.add("9");plusChoice.add("10");plusChoice.add("11");
	plusChoice.add("12");plusChoice.add("13");plusChoice.add("14");
	plusChoice.add("15");plusChoice.add("16");plusChoice.add("17");
	plusChoice.add("18");plusChoice.add("19");plusChoice.add("20");
	plusStrokesPanel.add(plusChoice);
	plusChoice.addItemListener(new ItemListener() {
	    public void itemStateChanged(ItemEvent e) {
		short strokes = (short)Integer.parseInt((String) e.getItem());
		if (strokes == plusStrokes) return;
		plusStrokes = strokes;
		KanjiInput.this.radStrokeQuery(plusStrokes);
	    }
	});
 	plusStrokesPanel.add(new Label(" additional strokes."));
	Button goButton = new Button("Go");
	goButton.addActionListener(new ActionListener(){
	    public void actionPerformed(ActionEvent e){
		short strokes = (short)Integer.parseInt
		    (plusChoice.getSelectedItem());
		if (plusStrokes == strokes) return;
		plusStrokes = strokes;
		KanjiInput.this.radStrokeQuery(plusStrokes);
	    }
	});
	plusStrokesPanel.add(goButton);
	this.add(plusStrokesPanel);
	this.pack();
	this.validate();
	this.requestFocus();
	
   }
    Label kanjiSelection;
    KselectPanel kselect;
    Button insertButton;
    Button lookupButton;
    
    void radStrokeQuery(short plusStrokes){
	short radical = RadTable.radNumber(radStrokes, radChar);
	String[] results = JJDic.doKdicRadStrokeQuery(radical, (short)(radStrokes+plusStrokes));
	Panel buttonPanel = new Panel();
	insertButton = new Button("Insert");
	insertButton.setEnabled(false);
	insertButton.addActionListener(new InsertListener());
	lookupButton = new Button("Lookup");
	lookupButton.setEnabled(false);
	lookupButton.addActionListener(new LookupListener());
	kanjiSelection = new Label("Selected kanji: none");
	buttonPanel.add(kanjiSelection);
	buttonPanel.add(insertButton);
	buttonPanel.add(lookupButton);
	if (kanjiDialog != null){
	    kanjiDialog.dispose();
	}
	kanjiDialog = new Dialog((Frame)this.getParent(),
				 "Radical "+displayChar+" plus "+plusStrokes+" strokes.");
	kanjiDialog.addWindowListener(new WindowAdapter(){
	    public void windowClosing(WindowEvent e){
		kanjiDialog.dispose();
	    }
	});
	kanjiDialog.add(buttonPanel, "South");
	kselect = new KselectPanel(results);
	kanjiDialog.add(kselect,"Center");
	kanjiDialog.pack();
	kanjiDialog.show();
	kanjiDialog.requestFocus();
    }
    class InsertListener implements ActionListener {
	public void actionPerformed(ActionEvent e){
	    /*
	     * This just appends it does not properly insert at the caret position.
	     * When textcomponents are fixed then I will change it.
	     */
	    KanjiInput.this.where.setText
		(KanjiInput.this.where.getText()+
		 KanjiInput.this.kselect.getSelection().charAt(0));
	    ((Window)KanjiInput.this.where.getParent()).requestFocus();
	}
    }
    class LookupListener implements ActionListener {
	Dialog kd;
	public void actionPerformed(ActionEvent e){
	    String ks = KanjiInput.this.kselect.getSelection();
	    if (kd != null) kd.dispose();
	    kd = new Dialog((Frame)KanjiInput.this.getParent(), ks.substring(0,1));
	    kd.add(new Kanji(ks),"Center");
	    kd.addWindowListener(new WindowAdapter(){
		public void windowClosing(WindowEvent e){
		    kd.dispose();
		}
	    });
	    kd.pack();
	    kd.show();
	    kd.requestFocus();
	}
    }
    
    class KselectPanel extends KanjiPanel {
	String selection = null;
	public KselectPanel(String[] entries){
	    super(entries);
	}
	public void kanjiSelection(String kSelection) {
	    selection = kSelection;
	    KanjiInput.this.kanjiSelection.setText("Selected kanji: "+selection.charAt(0));
	    KanjiInput.this.lookupButton.setEnabled(true);
	    KanjiInput.this.insertButton.setEnabled(true);
	    KanjiInput.this.kanjiDialog.validate();
	    KanjiInput.this.kanjiDialog.show();
	    KanjiInput.this.kanjiDialog.requestFocus();
	}
	String getSelection() {
	    return(selection);
	}
    }
    
}
