/*
 * This program is Copyright 1998 by Henry Cohen commercial rights are reserved.
 * The program is provided as is with no warranties of any sort including 
 * fitness for any purpose.  In no case will the author be held liable for
 * damages related to the use of this program.

 * The author hereby grants rights to use, copy, modify and create derived
 * products for individual, non-commercial purposes provided a copy of this
 * notice is included.  Any commercial use requires the written approval of 
 * the author.
 * Henry Cohen 1998.
 */

/*
 * Class KanjiPanel:  Display a list of kanji as a table
 * the class supports click selection of an individual character.
 * The layout is fairly simple.  All characters are layed out in
 * a square.  If the number of characters to be displayed is less than
 * a square number then some rows may be unfilled or not present.
 */
package jdic;

import java.awt.*;
import java.awt.event.*;

public class KanjiPanel extends Canvas {

    String[] klist;
    int N;                                     // Dimension of the display square
    int rows;                                  // Number of rows with characters
    int cellHeight, cellWidth, baseXOffset, baseYOffset; // Graphic dimensions
    Dimension size;
    final int BORDER = 3;
    Point mousePoint;
    private Point outerPoint = new Point(-100,-100); // for mouse exited events
    private int lastCellx = -1, lastCelly = -1;  // previous cell visited
    int mouseType;    // event type for mouse events
    public FontMetrics fontMetrics;
        
    public KanjiPanel(String[] kdictEntries) {

	
	this.klist = kdictEntries;
	/*
	 * Empty list case
	 */
	if ((kdictEntries == null)||(kdictEntries.length == 0)) {
	    size = new Dimension(0,0);
	    return;
	}
	/*
	 * Compute the number of rows and columns
	 */
	N = (int)Math.ceil(Math.sqrt((double)kdictEntries.length));
	rows = (int)Math.ceil((float)kdictEntries.length/(float)N);
	/*
	 * Use a big font for the button labels
	 */
	setFont(new Font("Dialog",Font.BOLD, 24));
	fontMetrics = this.getFontMetrics(this.getFont());
	/*
	 * Cell size is a function of the font used for drawing the labels
	 */
	cellHeight = fontMetrics.getMaxAscent()
                       +fontMetrics.getMaxDescent()
                       +fontMetrics.getLeading();
	cellWidth = fontMetrics.getMaxAdvance()+fontMetrics.getLeading();
	cellHeight = Math.max(cellHeight, cellWidth);
	cellWidth = cellHeight;
	/*
	 * And the panel size is a function of the cell size
	 */
	size = new Dimension();
	size.width = N*cellWidth + 2*BORDER+1;
	size.height = rows*cellHeight + 2*BORDER+1;
	/*
	 * Text baseline offset within the button for the label
	 */
	baseYOffset = fontMetrics.getMaxAscent()+4;
	baseXOffset = 4;
	/*
	 * Add listeners for mouse events.
	 * Everything but clicked since we do both pressed and released.
	 * [soap] I would like to discard extraneous mouse points here
	 * but the necessary component method contains() is not available
	 * from the inner class.  Therefore I have to do it in update()
	 */
	addMouseMotionListener(new MouseMotionListener () {
	    public void mouseMoved (MouseEvent scat) {
		mousePoint = scat.getPoint();
		mouseType = MouseEvent.MOUSE_MOVED;
		repaint();
	    }
	    public void mouseDragged (MouseEvent scat) {
		mousePoint = scat.getPoint();
		mouseType = MouseEvent.MOUSE_DRAGGED;
		repaint();
	    }
	});
	addMouseListener (new MouseAdapter () {
	    public void mousePressed (MouseEvent scat) {
		mousePoint = scat.getPoint();
		mouseType = MouseEvent.MOUSE_PRESSED;
		if (mousePoint.x < 0 || mousePoint.y<0) return ;
		repaint();
	    }
	    public void mouseReleased (MouseEvent scat) {
		mousePoint = scat.getPoint();
		mouseType = MouseEvent.MOUSE_RELEASED;
		if (mousePoint.x < 0 || mousePoint.y<0) return ;
		repaint();
		int x = ((mousePoint.x - BORDER)/cellWidth);   // first the column
		int y = ((mousePoint.y - BORDER)/cellHeight);  // and row 
		int cindex = y*N+x;  // The character index
                if (((0 <= cindex) && (cindex < klist.length))
		       && ((0<=x)&&(x<N))
		       && ((0<=y)&&(y<=rows))) {
		    /*
		     * This is a bit of Caffine! Look out for overloading    
		     * When this class is used for kanji input this method
		     * should be overridden.
		     */
		     kanjiSelection(klist[cindex]); 
		}
	    }
	    /*
	     * On an exit we should un-highlight the last cell
	     * and not redraw the current cell.
	     */
	    public void mouseExited(MouseEvent scat){
		mousePoint = outerPoint;
		mouseType = MouseEvent.MOUSE_EXITED;
		repaint();
	    }
	    /*
	     * On an entry we should highlight the current cell
	     * but not do anything to the last cell.
	     */
	    public void mouseEntered(MouseEvent scat){
		mousePoint = scat.getPoint();
		lastCellx = -200;
		lastCelly = -200;
		repaint();
	    }
	});
    }
    /*
     * This is a very simplistic approach to sizing the component.
     * I expect that I should not need more than 300 characters in
     * a panel and they should all fit on the screen at once.  If I
     * find that they don't fit then I may need to do better than this so
     * that a KanjiPanel can be put inside a scrolled pane.  The ScrollPane
     * component of the AWT is probably not sufficient though since this
     * pane should scroll by cells not by pixels.
     */
    public Dimension getMinimumSize() {
	return size;
    }
    public Dimension getMaximumSize() {
	return size;
    }
    public Dimension getPreferredSize(){
	return size;
    }
    public Dimension getSize() {
	return size;
    }
    /*
     * The paint method draws the entire figure.
     * Update only draws the part of the figure pointed to by the mouse.
     */
    public void paint (Graphics g){
	int cindex = 0;              // character in cell 
	
	/*
	 * Clear the entire background
	 * This is necessary because I replace the update() method with
	 * one that only redraws the subsection pointed to by the mouse.
	 * Paint is used to draw the entire figure after an expose event
	 * or when first made visible.
	 */
	Rectangle r = g.getClipBounds();
	g.setColor(this.getBackground());
	g.fillRect(r.x, r.y, r.height, r.width);
	
	/*
	 * Draw a bounding box
	 */
	g.draw3DRect(0, 0, size.width - 1, size.height - 1, true);
        g.draw3DRect(BORDER, BORDER, size.width - 2*BORDER -1, size.height - 2*BORDER -1, false);
	/*
	 * The following loops have the x and y coordinate computations
	 * hoisted to convert multiplication into addition and to avoid
	 * re-computing common subexpressions.  [Soap] The compiler should
	 * do this but I don't think that it does.
	 */

	int y = BORDER;
	for (int row = 0; row< rows; row++, y += cellHeight) {
	    for (int column = 0, x = BORDER
		     ; column < N && cindex < klist.length
		     ; column++, x += cellWidth) {
		g.setColor(this.getBackground());
		g.fill3DRect(x, y, cellWidth-1,cellHeight-1, true);
		g.setColor(this.getForeground());
		
		paintGlyph(g, klist[cindex++].substring(0,1), x, y);
		
	    }
	}
    }
    /*
     * redraw the previous cell pointed to by the mouse,
     * and/or the current cell pointed to.
     */ 
    public void update (Graphics g) {
	/*
	 * Compute the cell pointed to
	 * There seems to be a bug in the AWT somewhere because I get mouse
	 * events from outside of this component that is listeneing for them.
	 * i.e. although entered and exited events are correctly clipped
	 * move drag and button press events are clipped to the container.
	 * Because of this I have to check to be sure that we really are within
	 * the bounds of this component.
	 */
	int x = ((mousePoint.x - BORDER)/cellWidth);   // first the column
	int y = ((mousePoint.y - BORDER)/cellHeight);  // and row 
	int cindex = y*N+x;  // The character index
	
	/*
	 * Throw away any moved or dragged events in the same cell as last time.
	 * This decimates the number of events processed. The reduction could be
	 * as much as 10 or 20 to 1.
	 */
	if ((x == lastCellx) && (y == lastCelly) 
	    && ((mouseType == MouseEvent.MOUSE_MOVED)
	    ||(mouseType == MouseEvent.MOUSE_DRAGGED))) {
	    return;
	}

	/*
	 * Redraw the last cell visited in normal mode. 
	 * i.e. as a raised 3d rectangle.
	 * Compute the last cell indices.
	 */
	int lx = lastCellx*cellWidth + BORDER;
	int ly = lastCelly*cellHeight + BORDER;
	int lcindex = lastCelly*N+lastCellx;
	/*
	 * Verify that it is a valid cell.  It might not be if we just entered
	 * the grid.
	 */
	if ((0<=lcindex)&&(lcindex<klist.length) 
	    && ((0<=lastCellx)&&(lastCellx<N))
	       && ((0<=lastCelly)&&(lastCelly<=rows))) {
	    /*
	     * OK draw.
	     */
	    g.setColor(this.getBackground());
	    g.fillRect(lx, ly, cellWidth-1,cellHeight-1);
	    g.fill3DRect(lx, ly, cellWidth-1,cellHeight-1,true);
	    g.setColor(this.getForeground());
	    paintGlyph(g, klist[lcindex].substring(0,1), lx, ly);
	}
	lastCellx = x;
	lastCelly = y;

	/*
	 * Test for a valid cell. 
	 * The x and y conditions test for being within the defined rectangle
	 * and the cindex test verifies that the cell is really defined.
	 */
	if (((0<=cindex)&&(cindex<klist.length))
	    && ((0<=x)&&(x<N))
	    && ((0<=y)&&(y<=rows))) {
	    /*
	     * Draw the current cell as a depressed rectangle.
	     */
	    x = x*cellWidth + BORDER;  // now convert columns to pixels
	    y = y*cellHeight + BORDER; // and rows to pixels
	    g.setColor(this.getBackground());
	    g.fillRect(x, y, cellWidth, cellHeight);
	    g.fill3DRect(x, y, cellWidth-1,cellHeight-1,false);
	    /*
	     * If the Mouse is pressed then draw the label in a light color
	     * against a dark background until the button is released.  
	     * If the mouse is dragged then draw in grey until it is released.
	     * The selected character is the one in which the button is released.
	     */
	    if ((mouseType == MouseEvent.MOUSE_PRESSED) 
		|| (mouseType == MouseEvent.MOUSE_DRAGGED)) {
		g.setColor(this.getBackground());
	    } else {
		g.setColor(this.getForeground());
	    }
	    paintGlyph(g,klist[cindex].substring(0,1), x, y);
	}
    }
    /*
     * This method must be overridden when subclassed to achieve any
     * interresting behavior.
     */
    public void kanjiSelection(String kdictEntry) {
	System.err.println("kanjiSelection :"+kdictEntry);
    }
    /*
     * Method painGlyph.  I define this method so that I can subclass this object
     * and draw glyphs differently for radical/stroke lookups when the radical is
     * not available in the current font.  The coordinate (x,y) is the origin of
     * the cell not the character baseline.  This is so that subclasses can eaisly 
     * manipulate the cell background.
     */
    void paintGlyph(Graphics g, String glyph, int x, int y) {
	g.drawString(glyph, x+baseXOffset, y+baseYOffset);
    }
    /*
     * This main program is for testing purposes only.  
     * Since Java allows multiple mains in a package I won't take it out
     * but it doesn't do anything very interesting.
     */
    public static void main (String[] argv){
        String arg[] = new String[5];
	arg[0] = "0";
	arg[1] = "1";
	arg[2] = "2";
	arg[3] = "3";
	arg[4] = "4";
	KanjiPanel kp = new KanjiPanel(arg);
	Frame f = new Frame("test");
	
	f.addWindowListener(new WindowAdapter(){
	    public void windowClosing(WindowEvent e){
		System.exit(1);
	    }
	});
	Dialog d = new Dialog(f);
	d.addWindowListener(new WindowAdapter(){
	    public void windowClosing(WindowEvent e){
		System.exit(1);
	    }
	});
	d.add(kp);
	d.pack();
	d.setVisible(true);
    }
}



